import featureflow as ff
import unittest2
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import UnitNorm, PreprocessingPipeline
from learn import KMeans
from zounds.timeseries import ConstantRateTimeSeries, Seconds
from zounds.spectral import \
    TimeFrequencyRepresentation, FrequencyBand, LinearScale


class UnitNormTests(unittest2.TestCase):

    def get_model(self):
        @simple_in_memory_settings
        class Model(ff.BaseModel):

            unitnorm = ff.PickleFeature(
                UnitNorm,
                store=False)

            kmeans = ff.PickleFeature(
                KMeans,
                centroids=10,
                needs=unitnorm,
                store=False)

            pipeline = ff.PickleFeature(
                PreprocessingPipeline,
                needs=(unitnorm, kmeans),
                store=True)

        return Model

    def invert_and_assert_class(self, data):
        training = np.random.random_sample((100, 30))
        Model = self.get_model()
        _id = Model.process(unitnorm=training)
        model = Model(_id)
        transformed = model.pipeline.transform(data)
        inverted = transformed.inverse_transform()
        self.assertEqual(data.__class__, inverted.__class__)
        return inverted

    def test_inversion_returns_array(self):
        data = np.random.random_sample((10, 30))
        self.invert_and_assert_class(data)

    def test_inversion_returns_time_series(self):
        data = np.random.random_sample((33, 30))
        ts = ConstantRateTimeSeries(data, frequency=Seconds(1))
        inverted = self.invert_and_assert_class(ts)
        self.assertEqual(Seconds(1), inverted.frequency)

    def test_inversion_returns_time_frequency_representation(self):
        data = np.random.random_sample((33, 30))
        scale = LinearScale(FrequencyBand(20, 20000), 30)
        tf = TimeFrequencyRepresentation(
            data,
            frequency=Seconds(1),
            duration=Seconds(2),
            scale=scale)
        inverted = self.invert_and_assert_class(tf)
        self.assertEqual(Seconds(1), inverted.frequency)
        self.assertEqual(Seconds(2), inverted.duration)
        self.assertEqual(scale, inverted.scale)

    def test_can_easily_get_at_codebook(self):
        # KLUDGE: This doesn't belong here, but putting it here is convenient
        # right now
        training = np.random.random_sample((100, 30))
        Model = self.get_model()
        _id = Model.process(unitnorm=training)
        model = Model(_id)
        self.assertIsInstance(model.pipeline[-1].codebook, np.ndarray)



from __future__ import division
import numpy as np
import unittest2
from tfrepresentation import TimeFrequencyRepresentation
from frequencyscale import LinearScale, LogScale, FrequencyBand
from weighting import AWeighting
from zounds.timeseries import Seconds


class TimeFrequencyRepresentationTests(unittest2.TestCase):

    def test_from_example(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.ones((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        from_example = TimeFrequencyRepresentation.from_example(
                np.ones((30, 100)), tf)
        self.assertEqual(tf.shape, from_example.shape)
        self.assertEqual(tf.frequency, from_example.frequency)
        self.assertEqual(tf.duration, from_example.duration)
        self.assertEqual(tf.scale, from_example.scale)

    def test_can_multiply_by_frequency_weighting_linear_scale(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.ones((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        result = tf * AWeighting()
        self.assertIsInstance(result, TimeFrequencyRepresentation)
        peak_frequency_band = FrequencyBand(9000, 11000)
        lower_band = FrequencyBand(100, 300)
        peak_slice = np.abs(result[:, peak_frequency_band]).max()
        lower_slice = np.abs(result[:, lower_band]).max()
        self.assertGreater(peak_slice, lower_slice)

    def test_can_multiply_by_frequency_weighting_log_scale(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LogScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.ones((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        result = tf * AWeighting()
        self.assertIsInstance(result, TimeFrequencyRepresentation)
        peak_frequency_band = FrequencyBand(9000, 11000)
        lower_band = FrequencyBand(100, 300)
        peak_slice = np.abs(result[:, peak_frequency_band]).max()
        lower_slice = np.abs(result[:, lower_band]).max()
        self.assertGreater(peak_slice, lower_slice)

    def test_can_multiply_by_array(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.ones((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        result = tf * np.ones(100)
        self.assertIsInstance(result, TimeFrequencyRepresentation)
        np.testing.assert_allclose(tf, result)

    def test_can_convert_to_string(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        s = str(tf)
        self.assertTrue(len(s))

    def test_can_use_single_integer_index(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        indexed = tf[0]
        self.assertEqual((100,), indexed.shape)
        self.assertIsInstance(indexed, TimeFrequencyRepresentation)

    def test_can_use_list_of_integers_as_index(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        indexed = tf[[0, 10, 14]]
        self.assertEqual((3, 100), indexed.shape)
        self.assertIsInstance(indexed, TimeFrequencyRepresentation)

    def test_can_construct_instance(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        self.assertIsInstance(tf, TimeFrequencyRepresentation)

    def test_raises_if_scale_length_does_not_match_frequency_dimension(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 1000)
        self.assertRaises(ValueError, lambda: TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale))

    def test_array_must_be_at_least_2d(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        self.assertRaises(ValueError, lambda: TimeFrequencyRepresentation(
                np.zeros(30),
                frequency=frequency,
                duration=duration,
                scale=scale))

    def test_can_slice_frequency_dimension_with_integer_indices(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        sliced = tf[:, 10: 20]
        self.assertEqual((30, 10), sliced.shape)
        self.assertIsInstance(sliced, TimeFrequencyRepresentation)

    def test_can_slice_frequency_dimensions_with_frequency_band(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        bands = list(scale)
        sliced = tf[:, bands[0]]
        self.assertEqual((30, 1), sliced.shape)
        self.assertIsInstance(sliced, TimeFrequencyRepresentation)

    def test_can_slice_freq_dimension_with_freq_band_spanning_bins(self):
        frequency = Seconds(1)
        duration = Seconds(1)
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        tf = TimeFrequencyRepresentation(
                np.zeros((30, 100)),
                frequency=frequency,
                duration=duration,
                scale=scale)
        bands = list(scale)
        wide_band = FrequencyBand(bands[0].start_hz, bands[9].stop_hz)
        sliced = tf[:, wide_band]
        self.assertEqual((30, 10), sliced.shape)
        self.assertIsInstance(sliced, TimeFrequencyRepresentation)



import unittest2
import numpy as np
from weighting import AWeighting
from tfrepresentation import TimeFrequencyRepresentation
from frequencyscale import LinearScale, FrequencyBand
from zounds.timeseries import Seconds


class WeightingTests(unittest2.TestCase):
    def test_can_get_weights_from_tf_representation(self):
        t = Seconds(1)
        tf = TimeFrequencyRepresentation(
            np.ones((90, 100)),
            frequency=t,
            duration=t,
            scale=LinearScale(FrequencyBand(20, 22050), 100))
        weighting = AWeighting()
        weights = weighting.weights(tf)
        self.assertEqual((100,), weights.shape)

    def test_can_get_weights_from_scale(self):
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        weighting = AWeighting()
        weights = weighting.weights(scale)
        self.assertEqual((100,), weights.shape)



from sliding_window import \
    SlidingWindow, IdentityWindowingFunc, OggVorbisWindowingFunc
from zounds.timeseries import \
    AudioSamples, SR22050, SR44100, SR11025, SR48000, SR96000, SampleRate, \
    Seconds, Milliseconds, ConstantRateTimeSeriesFeature
from zounds.util import simple_in_memory_settings
from zounds.basic import resampled
from zounds.synthesize import NoiseSynthesizer
import numpy as np
import unittest2


class WindowingFunctionTests(unittest2.TestCase):
    def test_multiply_lhs(self):
        samples = np.random.random_sample(44)
        wf = IdentityWindowingFunc()
        np.testing.assert_allclose(wf * samples, samples)

    def test_multiply_rhs(self):
        samples = np.random.random_sample(33)
        wf = IdentityWindowingFunc()
        np.testing.assert_allclose(samples * wf, samples)


class OggVorbisWindowingFunctionTests(unittest2.TestCase):
    def test_multilpy_many_frames(self):
        samples = np.random.random_sample((10, 3))
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)

    def test_multiply_one_frame(self):
        samples = np.random.random_sample((10, 1))
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)

    def test_multiply_1d(self):
        samples = np.random.random_sample(10)
        wf = OggVorbisWindowingFunc()
        result = wf * samples
        self.assertEqual(samples.shape, result.shape)


class SlidingWindowTests(unittest2.TestCase):
    def _check(self, samplerate, expected_window_size, expected_step_size):
        sw = SlidingWindow(wscheme=samplerate.half_lapped())
        samples = AudioSamples(
                np.zeros(5 * samplerate.samples_per_second), samplerate)
        sw._enqueue(samples, None)
        self.assertEqual(expected_window_size, sw._windowsize)
        self.assertEqual(expected_step_size, sw._stepsize)

    def test_correct_window_and_step_size_at_96000(self):
        self._check(SR96000(), 4096, 2048)

    def test_correct_window_and_step_size_at_48000(self):
        self._check(SR48000(), 2048, 1024)

    def test_correct_window_and_step_size_at_22050(self):
        self._check(SR22050(), 1024, 512)

    def test_correct_window_and_step_size_at_44100(self):
        self._check(SR44100(), 2048, 1024)

    def test_correct_window_and_step_size_at_11025(self):
        self._check(SR11025(), 512, 256)

    def test_can_persist_and_retrieve_with_second_long_windowing_scheme(self):
        samplerate = SR22050()
        rs = resampled(resample_to=samplerate)

        window_size = Seconds(1)
        wscheme = SampleRate(window_size, window_size)

        @simple_in_memory_settings
        class Document(rs):
            windowed = ConstantRateTimeSeriesFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    needs=rs.resampled,
                    store=True)

        synth = NoiseSynthesizer(samplerate)
        audio = synth.synthesize(Milliseconds(5500))

        _id = Document.process(meta=audio.encode())
        doc = Document(_id)

        self.assertEqual(6, len(doc.windowed))



import numpy as np
import featureflow as ff
from zounds.timeseries import \
    ConstantRateTimeSeries, ConstantRateTimeSeriesMetadata
import frequencyscale


class TimeFrequencyRepresentation(ConstantRateTimeSeries):
    """
    A class that encapsulates time-frequency representation data.  The first
    axis represents time, the second axis represents frequency, and any
    subsequent axes contain multidimensional data about time-frequency positions
    """

    def __new__(cls, arr, frequency=None, duration=None, scale=None):
        if len(arr.shape) < 2:
            raise ValueError('arr must be at least 2D')

        if len(scale) != arr.shape[1]:
            raise ValueError('scale must have same size as dimension 2')

        obj = ConstantRateTimeSeries.__new__(cls, arr, frequency, duration)
        obj.scale = scale
        return obj

    def kwargs(self, **kwargs):
        return super(TimeFrequencyRepresentation, self).kwargs(
                scale=self.scale, **kwargs)

    @classmethod
    def from_example(cls, arr, example):
        return cls(
                arr,
                frequency=example.frequency,
                duration=example.duration,
                scale=example.scale)

    def __array_finalize__(self, obj):
        super(TimeFrequencyRepresentation, self).__array_finalize__(obj)
        if obj is None:
            return
        self.scale = getattr(obj, 'scale', None)

    def _freq_band_to_integer_indices(self, index):
        if not isinstance(index, frequencyscale.FrequencyBand):
            return index

        return self.scale.get_slice(index)

    def __getitem__(self, index):
        try:
            slices = map(self._freq_band_to_integer_indices, index)
        except TypeError:
            slices = self._freq_band_to_integer_indices(index)
        return super(TimeFrequencyRepresentation, self).__getitem__(slices)


class TimeFrequencyRepresentationMetaData(ConstantRateTimeSeriesMetadata):
    def __init__(
            self,
            dtype=None,
            shape=None,
            frequency=None,
            duration=None,
            scale=None):
        super(TimeFrequencyRepresentationMetaData, self).__init__(
                dtype=dtype,
                shape=shape,
                frequency=frequency,
                duration=duration)
        self.scale = self._decode_scale(scale)

    @staticmethod
    def from_time_frequency_representation(tf):
        return TimeFrequencyRepresentationMetaData(
                dtype=tf.dtype,
                shape=tf.shape[1:],
                frequency=tf.frequency,
                duration=tf.duration,
                scale=tf.scale)

    def _encode_scale(self, scale):
        return \
            scale.__class__.__name__, \
            (scale.start_hz, scale.stop_hz), \
            scale.n_bands

    def _decode_scale(self, t):
        if isinstance(t, frequencyscale.FrequencyScale):
            return t
        cls_name, span, n_bands = t
        scale_cls = getattr(frequencyscale, cls_name)
        return scale_cls(frequencyscale.FrequencyBand(*span), n_bands)

    def __repr__(self):
        return repr((
            str(np.dtype(self.dtype)),
            self.shape,
            self._encode_timedelta(self.frequency),
            self._encode_timedelta(self.duration),
            self._encode_scale(self.scale)
        ))


class TimeFrequencyRepresentationEncoder(ff.NumpyEncoder):
    def __init__(self, needs=None):
        super(TimeFrequencyRepresentationEncoder, self).__init__(needs=needs)

    def _prepare_data(self, data):
        return data

    def _prepare_metadata(self, data):
        return TimeFrequencyRepresentationMetaData\
            .from_time_frequency_representation(data)


class TimeFrequencyRepresentationDecoder(ff.BaseNumpyDecoder):
    def __init__(self):
        super(TimeFrequencyRepresentationDecoder, self).__init__()

    def _unpack_metadata(self, flo):
        return TimeFrequencyRepresentationMetaData.unpack(flo)

    def _wrap_array(self, raw, metadata):
        return TimeFrequencyRepresentation(
            raw, metadata.frequency, metadata.duration, metadata.scale)


class TimeFrequencyRepresentationFeature(ff.Feature):
    def __init__(
            self,
            extractor,
            needs=None,
            store=False,
            key=None,
            encoder=TimeFrequencyRepresentationEncoder,
            decoder=TimeFrequencyRepresentationDecoder(),
            **extractor_args):
        super(TimeFrequencyRepresentationFeature, self).__init__(
            extractor,
            needs=needs,
            store=store,
            encoder=encoder,
            decoder=decoder,
            key=key,
            **extractor_args)



import numpy as np
import unittest2
from zounds.basic import resampled
from zounds.util import simple_in_memory_settings
from zounds.timeseries import \
    SR11025, SR22050, ConstantRateTimeSeriesFeature, Seconds, Picoseconds
from zounds.timeseries.samplerate import SampleRate, HalfLapped
from zounds.synthesize import SineSynthesizer, DCTIVSynthesizer, MDCTSynthesizer
from zounds.spectral import SlidingWindow, DCTIV, MDCT
from tfrepresentation import \
    TimeFrequencyRepresentation, TimeFrequencyRepresentationFeature


class MDCTTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR11025()
        rs = resampled(resample_to=self.samplerate)

        wscheme = HalfLapped()

        @simple_in_memory_settings
        class Document(rs):
            windowed = ConstantRateTimeSeriesFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    needs=rs.resampled,
                    store=False)

            mdct = TimeFrequencyRepresentationFeature(
                    MDCT,
                    needs=windowed,
                    store=True)

        ss = SineSynthesizer(self.samplerate)
        self.audio = ss.synthesize(Seconds(5), [440., 660., 880.])

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    def test_is_correct_type(self):
        self.assertIsInstance(self.doc.mdct, TimeFrequencyRepresentation)

    def test_has_correct_nyquist_frequency(self):
        self.assertEqual(self.samplerate.nyquist, self.doc.mdct.scale.stop_hz)

    def test_reconstruction(self):
        ds = MDCTSynthesizer()
        recon = ds.synthesize(self.doc.mdct)

        # self.assertEqual(len(self.audio), len(recon))
        self.assertEqual(self.audio.samplerate, recon.samplerate)
        # orig_fft = abs(np.fft.rfft(self.audio))
        # recon_fft = abs(np.fft.rfft(recon))
        # orig_peaks = set(np.argsort(orig_fft)[-3:])
        # recon_peaks = set(np.argsort(recon_fft)[-3:])
        # print len(orig_peaks), len(recon_peaks)
        # # ensure that the original and reconstruction have the same three
        # # spectral peaks
        # self.assertEqual(orig_peaks, recon_peaks)


class DCTIVTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR22050()
        rs = resampled(resample_to=self.samplerate)

        window_size = Picoseconds(int(1e12))
        wscheme = SampleRate(window_size, window_size)

        @simple_in_memory_settings
        class Document(rs):
            windowed = ConstantRateTimeSeriesFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    needs=rs.resampled,
                    store=False)

            dct = TimeFrequencyRepresentationFeature(
                    DCTIV,
                    needs=windowed,
                    store=True)

        ss = SineSynthesizer(self.samplerate)
        self.audio = ss.synthesize(Seconds(5), [440., 660., 880.])

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    def test_is_correct_type(self):
        self.assertIsInstance(self.doc.dct, TimeFrequencyRepresentation)

    def test_has_correct_nyquist_frequency(self):
        self.assertEqual(self.samplerate.nyquist, self.doc.dct.scale.stop_hz)

    def test_reconstruction(self):
        ds = DCTIVSynthesizer()
        recon = ds.synthesize(self.doc.dct)

        self.assertEqual(len(self.audio), len(recon))
        self.assertEqual(self.audio.samplerate, recon.samplerate)
        orig_fft = abs(np.fft.rfft(self.audio))
        recon_fft = abs(np.fft.rfft(recon))
        orig_peaks = set(np.argsort(orig_fft)[-3:])
        recon_peaks = set(np.argsort(recon_fft)[-3:])
        # ensure that the original and reconstruction have the same three
        # spectral peaks
        self.assertEqual(orig_peaks, recon_peaks)



import unittest2
import featureflow as ff
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import Slicer, PreprocessingPipeline


class SlicerTests(unittest2.TestCase):
    def get_model(self, slicex):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            sliced = ff.PickleFeature(
                    Slicer,
                    slicex=slicex,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(sliced,),
                    store=True)

        return Model

    def test_can_slice_array_with_slice_instance(self):
        training = np.ones((100, 30))
        Model = self.get_model(slicex=slice(10, 20))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = np.ones((100, 30))
        transformed = model.pipeline.transform(data)
        self.assertEqual((100, 10), transformed.data.shape)
        inverted = transformed.inverse_transform()
        self.assertEqual((100, 30), inverted.shape)
        np.testing.assert_allclose(transformed.data, 1)
        np.testing.assert_allclose(inverted[:, :10], 0)
        np.testing.assert_allclose(inverted[:, 20:], 0)

    def test_can_slice_multi_dimensional(self):
        training = np.ones((100, 30, 50))
        Model = self.get_model(slicex=slice(10, 20))
        _id = Model.process(sliced=training)
        model = Model(_id)
        data = np.ones((100, 30, 50))
        transformed = model.pipeline.transform(data)
        self.assertEqual((100, 30, 10), transformed.data.shape)
        inverted = transformed.inverse_transform()
        self.assertEqual((100, 30, 50), inverted.shape)
        np.testing.assert_allclose(transformed.data, 1)
        np.testing.assert_allclose(inverted[..., :10], 0)
        np.testing.assert_allclose(inverted[..., 20:], 0)